import mne
import numpy as np
from scipy import stats


subject_names = ['s129','s240','s241','s242','s243','s247','s248','s300']

evoked_list = []
inv_list = []
for name in subject_names:

    ############# LOAD EPOCHS ######################
    name_epoch = '{}-epochs.fif'.format(name)
    fname_epochs = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/'.format(name) + name_epoch
    epochs = mne.read_epochs(fname_epochs)
    evoked = epochs.average()
    evoked_list.append(evoked)

    name_inverse = '{}-inv.fif'.format(name)
    fname_inverse = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/'.format(name) + name_inverse
    inverse_operator = mne.minimum_norm.read_inverse_operator(fname_inverse)
    inv_list.append(inverse_operator)

def average_between(evoked,tstart,tstop):
    evoked_out = evoked.copy()
    time = evoked_out.times
    start = (time < tstart+0.0005) * (time > tstart-0.0005)
    stop = (time < tstop+0.0005) * (time > tstop-0.0005)
    t0 = np.where(time == 0)[0][0]
    #print(start,stop)
    start_idx = np.where(start)[0][0]
    stop_idx = np.where(stop)[0][0]
    data_arr = []
    for i,value in enumerate(evoked.data[:][:]):
        data = value[start_idx:stop_idx]
        data_mean = data.mean()
        data_arr.append(data_mean)
    data_arr =  np.array(data_arr)
    data_arr = data_arr.reshape(len(data_arr),1)
    evoked_out.data = data_arr
    evoked_out.times = time[t0:t0+1]
    return evoked_out
#evoked_test = average_between(evoked_list[0],-0.150,-0.020)


def get_stcs_and_data(evoked_list,inv_list):
    stcs_baseline = []
    stcs_pre_rem = []
    stcs_data_baseline = []
    stcs_data_pre_rem = []
    method = 'sLORETA'  # 'sloreta' "dSPM" “MNE” | “dSPM” | “sLORETA” | “eLORETA”
    snr = 3.
    lambda2 = 1. / snr ** 2
    for evoked,inv in zip(evoked_list,inv_list):
        evoked_baseline = average_between(evoked,-0.1500,-0.140)
        evoked_pre_rem = average_between(evoked,-0.150,-0.020)
        stc_baseline = mne.minimum_norm.apply_inverse(evoked_baseline, inv, lambda2,
                        method=method, pick_ori=None)
        stcs_baseline.append(stc_baseline)
        stc_pre_rem = mne.minimum_norm.apply_inverse(evoked_pre_rem, inv, lambda2,
                                                      method=method, pick_ori=None)
        stcs_pre_rem.append(stc_pre_rem)
        stcs_data_baseline.append(stc_baseline.data)
        stcs_data_pre_rem.append(stc_pre_rem.data)

    data_b = np.array(stcs_data_baseline)
    data_pre = np.array(stcs_data_pre_rem)
    data_b = data_b.reshape(data_b.shape[0], data_b.shape[1])
    data_pre = data_pre.reshape(data_pre.shape[0], data_pre.shape[1])
    return stcs_baseline, stcs_pre_rem, data_b, data_pre

stcs_baseline, stcs_pre_rem, data_b, data_pre = get_stcs_and_data(evoked_list,inv_list)
# for i in range(len(stcs_baseline)):
#     stcs_pre_rem[i].plot()
#     stcs_baseline[i].plot()


t_statistic, pvalue = stats.ttest_rel(data_b, data_pre, axis=0)


statistical_significant_index = []
data_arr = []
for i in range(len(pvalue)):
    if pvalue[i] < 0.05:
        statistical_significant_index.append(i)
        data_arr.append(1)
    else:
        data_arr.append(0)
print(len(statistical_significant_index))
data_arr = np.array(data_arr)
data_arr = data_arr.reshape(len(data_arr),1)

stc_out_evoked_average = mne.SourceEstimate(data_arr,stcs_baseline[0].vertices,tmin=0,tstep=0)
subject = 'fsaverage'
parc = 'aparc'  # 'aparc_sub'

from matplotlib import cm
#Defining colomap
colors = cm.get_cmap('jet')
cmap_colors = colors(np.linspace(0, 1, 256))
cmap_colors = cmap_colors[:,0:3]
cmap = cm.viridis
cmap.colors = cmap_colors
for i in range(0,2):
    cmap.colors[i] = [0.9, 0.9, 0.9]

from mne.datasets import fetch_fsaverage
import os
from matplotlib import pyplot as plt
fs_dir = fetch_fsaverage(verbose=True)
subjects_dir = os.path.dirname(fs_dir)

brain = stc_out_evoked_average.plot(subject=subject,hemi='both',subjects_dir=subjects_dir,surface='pial',
            transparent=False,colormap=cmap,cortex='high_contrast',
            clim=dict(kind='value', lims=[0.99, 0.99, 1]),smoothing_steps=2)
brain.add_annotation(parc)


def get_stcs_for_ttest(evoked_list,inv_list):
    stcs = []
    method = 'sLORETA'  # 'sloreta' "dSPM" “MNE” | “dSPM” | “sLORETA” | “eLORETA”
    snr = 3.
    lambda2 = 1. / snr ** 2
    for evoked,inv in zip(evoked_list,inv_list):
        stcs.append(mne.minimum_norm.apply_inverse(evoked, inv, lambda2,method=method, pick_ori=None))
    stcs_baseline = []
    stcs_onset = []
    for stc in stcs:
        time = stc.times
        start2 = (time < -0.1495) * (time > -0.1505)
        stop2 = (time < -0.0195) * (time > -0.0205)
        start = (time < -0.1495) * (time > -0.1505)
        stop = (time < -0.1395) * (time > -0.1405)
        index_start = np.where(start)[0][0]
        index_stop = np.where(stop )[0][0]
        index_start2 = np.where(start2)[0][0]
        index_stop2 = np.where(stop2)[0][0]
        index_onset = np.where(time == 0)[0][0]
        data_arr_base = []
        data_arr_pre = []
        for value in stc.data[:]:
            data = value[index_start:index_stop]
            data_mean = data.mean()
            data_arr_base.append(data_mean)
            data_pre = value[index_start2:index_stop2]
            data_pre_mean = data_pre.mean()
            data_arr_pre.append(data_pre_mean)
        stcs_baseline.append(data_arr_base)
        stcs_onset.append(data_arr_pre)
    return np.array(stcs_baseline),np.array(stcs_onset)

stcs_b, stcs_o = get_stcs_for_ttest(evoked_list,inv_list)


t_statistic, pvalue = stats.ttest_rel(stcs_b, stcs_o, axis=0)


statistical_significant_index = []
data_arr = []
for i in range(len(pvalue)):
    if pvalue[i] < 0.05: # 0.001: gives 1060 points
        statistical_significant_index.append(i)
        data_arr.append(1)
    else:
        data_arr.append(0)
print(len(statistical_significant_index))
data_arr = np.array(data_arr)
data_arr = data_arr.reshape(len(data_arr),1)


stc_out_stc_averaged = mne.SourceEstimate(data_arr,stcs_baseline[0].vertices,tmin=0,tstep=0)
subject = 'fsaverage'
parc = 'aparc'  # 'aparc_sub'

from matplotlib import cm
#Defining colomap
colors = cm.get_cmap('jet')
cmap_colors = colors(np.linspace(0, 1, 256))
cmap_colors = cmap_colors[:,0:3]
cmap = cm.viridis
cmap.colors = cmap_colors
for i in range(0,2):
    cmap.colors[i] = [0.9, 0.9, 0.9]

from mne.datasets import fetch_fsaverage
import os
from matplotlib import pyplot as plt
fs_dir = fetch_fsaverage(verbose=True)
subjects_dir = os.path.dirname(fs_dir)

brain = stc_out_stc_averaged.plot(subject=subject,hemi='both',subjects_dir=subjects_dir,surface='pial',
            transparent=False,colormap=cmap,cortex='high_contrast',
            clim=dict(kind='value', lims=[0.99, 0.99, 1]),smoothing_steps=2)
brain.add_annotation(parc)

# brain = stc_out.plot(subject=subject,hemi='split',subjects_dir=subjects_dir,surface='pial',
#             transparent=False,colormap=cmap,cortex='high_contrast',
#             clim=dict(kind='value', lims=[0.99, 0.99, 1]),smoothing_steps=4)
# brain.add_annotation(parc)

#
# views = ['lateral', 'medial', 'rostral', 'caudal', 'dorsal', 'ventral',
#  'frontal', 'parietal', 'axial', 'sagittal', 'coronal']
# kwargs = dict(subject = subject, views=views[2],
#     hemi='both',clim=dict(kind='value', lims=[0, 0, 1]),
#     background='black',smoothing_steps=2,colorbar=False)
# brain = stc_out.plot(**kwargs) # show_traces=True ,dict(silhouette=True) show_toolbar= False,
# brain.add_annotation(parc)


############################# Samle stcs i en plot ##############################################
# stcs_baseline, stcs_pre_rem


# ############################## CONNECTIVITY ###########################################
# # Kan gjøres ved å finne nye stcs over tid og ikke med mean value
# # Download fsaverage files
# from mne.datasets import fetch_fsaverage
# import os
# from matplotlib import pyplot as plt
# fs_dir = fetch_fsaverage(verbose=True)
# subjects_dir = os.path.dirname(fs_dir)
# subject = 'fsaverage'
# parc = 'aparc'
#
# labels = mne.read_labels_from_annot(subject)
#
# labels_parc = [label for label in labels if label.name.lower() != 'unknown-lh' and label.name.lower() != 'unknown-rh'] ### remove unknown-lh
#
# # stcs_baseline, stcs_pre_rem
# src = inv_list[0]['src']
# # Average the source estimates within each label of the cortical parcellation
# # and each sub structures contained in the src space
# # If mode = 'mean_flip' this option is used only for the cortical label
# label_ts_baseline = mne.extract_label_time_course(stcs_baseline, labels_parc, src, mode='mean_flip',allow_empty=False, return_generator=False)
# label_ts_pre_rem = mne.extract_label_time_course(stcs_pre_rem, labels_parc, src, mode='mean_flip',allow_empty=False, return_generator=False)
#
# # We compute the connectivity in the alpha band and plot it using a circular
# # graph layout
# fmin = 8.
# fmax = 13.
# sfreq = 1024  # the sampling frequency
# con, freqs, times, n_epochs, n_tapers = mne.connectivity.spectral_connectivity(
#     label_ts_baseline, method='pli', mode='multitaper', sfreq=sfreq, fmin=fmin,
#     fmax=fmax, faverage=True, mt_adaptive=True, n_jobs=1)
# # fmin = 8.
# # fmax = 13.
# # sfreq = epochs.info['sfreq']  # the sampling frequency
# # con, freqs, times, n_epochs, n_tapers = mne.connectivity.spectral_connectivity(
# #     label_ts_pre_rem, method='pli', mode='multitaper', sfreq=sfreq, fmin=fmin,
# #     fmax=fmax, faverage=True, mt_adaptive=True, n_jobs=1)
#
#
# # We create a list of Label containing also the sub structures
# labels_aseg = mne.get_volume_labels_from_src(src, subject, subjects_dir)
# labels = labels_parc + labels_aseg
#
# # read colors
# node_colors = [label.color for label in labels]
#
# # We reorder the labels based on their location in the left hemi
# label_names = [label.name for label in labels]
# lh_labels = [name for name in label_names if name.endswith('lh')]
# rh_labels = [name for name in label_names if name.endswith('rh')]
#
# # Get the y-location of the label
# label_ypos_lh = list()
# for name in lh_labels:
#     idx = label_names.index(name)
#     ypos = np.mean(labels[idx].pos[:, 1])
#     label_ypos_lh.append(ypos)
# try:
#     idx = label_names.index('Brain-Stem')
# except ValueError:
#     pass
# else:
#     ypos = np.mean(labels[idx].pos[:, 1])
#     lh_labels.append('Brain-Stem')
#     label_ypos_lh.append(ypos)
#
#
# # Reorder the labels based on their location
# lh_labels = [label for (yp, label) in sorted(zip(label_ypos_lh, lh_labels))]
#
# # For the right hemi
# rh_labels = [label[:-2] + 'rh' for label in lh_labels
#              if label != 'Brain-Stem' and label[:-2] + 'rh' in rh_labels]
#
# # Save the plot order
# node_order = lh_labels[::-1] + rh_labels
#
# node_angles = mne.viz.circular_layout(label_names, node_order, start_pos=90,
#                               group_boundaries=[0, len(label_names) // 2])
#
#
# # Plot the graph using node colors from the FreeSurfer parcellation. We only
# # show the 300 strongest connections.
# conmat = con[:, :, 0]
# fig = plt.figure(num=None, figsize=(8, 8), facecolor='black')
# mne.viz.plot_connectivity_circle(conmat, label_names, n_lines=300,
#                          node_angles=node_angles, node_colors=node_colors,
#                          title='All-to-All Connectivity left-Auditory '
#                          'Condition (PLI)', fig=fig)
#
# ############################## CONNECTIVITY END ###########################################