import mne
import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

file_name = 's0248sleep.bdf'
subject_name = 's248'

data_path = os.path.expanduser("C:/Users/oledr/Desktop/Sleep Data 128ch") # REM from onedrive  s0129-Only-REM
raw_fname = os.path.join(data_path,file_name)# Activetwo2.bdf s0129sleep.bdf s0237sleep.bdf   namelist[0]
raw_full_size = mne.io.read_raw_bdf(raw_fname, preload=False) # False
sfreq = raw_full_size.info['sfreq']

tmin, tmax = 0, len(raw_full_size)/sfreq-0.1
sample_rate = 300 # sfreq



eve_name = '{}-eve.fif'.format(subject_name)
event_new_fname = "C:/Users/oledr/Desktop/Sleep Data 128ch/EM events/" + eve_name
events_from_file = mne.read_events(event_new_fname)



event_timings = events_from_file[:,0]
window_size = 30
window_list = []
for window in range(int(tmax/window_size)):
    EMs_in_window = np.where(np.logical_and(event_timings>=(window*window_size*1024), event_timings<=((window+1)*window_size*1024)))
    window_list.append(len(EMs_in_window[0]))

hight = max(window_list)
df = pd.read_csv('C:/Users/oledr/Desktop/Sleep Data 128ch/EM events/s0{}sleepstage_adaptation.csv'.format(subject_name[1:]))
window_score = df['category']
window_rem_score = []
for val in window_score:
    if val == 'REM':
        window_rem_score.append(hight+2)
    else:
        window_rem_score.append(0)

y = np.linspace(0,len(window_list)-1,num=len(window_list),dtype=int)
EMs_arr = []
y_arr = []
for i, value in enumerate(window_list):
    if value > 0:
        EMs_arr.append(value)
        y_arr.append(y[i])

plt.scatter(y_arr,EMs_arr,label='algorithm EMs per window')
plt.ylabel('EMs per window')
plt.xlabel('Window number, (window lenght = {} seconds)'.format(window_size))
plt.fill(window_rem_score,color = 'blue' ,label='windows scored REM',alpha=0.2)
plt.grid()
plt.show()
# plt.locator_params(axis='x', nbins=len(ems_in_rem)+1)

def rem_sleep_nr(rem_score):
    sleep_nr = 0
    rem_nr_index = []
    start = 0
    for i in range(len(rem_score)):
        if i < 1:
            continue
        if rem_score[i] > 0:
            if rem_score[i-1] == 0:
                sleep_nr += 1
                start = i
        else:
            if rem_score[i-1] > 0 :
                stop = i-1
                rem_nr_index.append([sleep_nr,start,stop])
    if rem_score[-1] > 0:
        rem_nr_index.append([sleep_nr,start,len(rem_score)-1])
    return rem_nr_index

def groupe_close(arr):
    grouped = []
    samle = []
    for i in range(len(arr)-1):
        if len(samle) < 1:
            samle.append(arr[i])
        if arr[i+1][1] - arr[i][2] < 10:
            samle.append(arr[i+1])
        else:
            grouped.append(samle)
            samle = []
    if len(samle) > 0:
        grouped.append(samle)
    return grouped

def Ems_in_rem_periode(rem_gruped,ems_in_window):
    ems_in_periode = []
    for groupe in rem_gruped:
        ems = 0
        for periode in groupe:
            ems += sum(ems_in_window[periode[1]:periode[2]+1])
        ems_in_periode.append(ems)
    return  ems_in_periode

def Em_density_in_rem_periode(rem_gruped,ems_in_window):
    ems_density_in_periode = []
    time_in_periode = []
    for groupe in rem_gruped:
        ems = 0
        time = 0
        for periode in groupe:
            ems += sum(ems_in_window[periode[1]:periode[2]+1])
            time += periode[2]+1 - periode[1]
        time_in_periode.append(time)
        ems_density_in_periode.append(2*ems/time)
    return  ems_density_in_periode,time_in_periode


REM_nrs = rem_sleep_nr(window_rem_score)
REM_gruped = groupe_close(REM_nrs)
ems_in_rem = Ems_in_rem_periode(REM_gruped,window_list)
em_density_gruped,time_in_periodes = Em_density_in_rem_periode(REM_gruped,window_list)



###### Plot EMS per REM #############
plt.figure()
plt.bar(list(range(1,len(ems_in_rem)+1)),ems_in_rem,label='algorithm EMs in REM nr ')
plt.ylabel('EMs per REM period')
plt.xlabel('REM period'.format(window_size))
plt.locator_params(axis='x', nbins=len(ems_in_rem)+1)
plt.show()


####### Plot EM density per REM #############
plt.figure()
plt.bar(list(range(1,len(em_density_gruped)+1)),em_density_gruped,label='algorithm EMs in REM nr ')
plt.ylabel('EMs density per REM period')
plt.xlabel('REM period'.format(window_size))
plt.locator_params(axis='x', nbins=len(ems_in_rem)+1)
plt.show()

############### save np array ####################################
# fname_np_array = 'C:/Users/oledr/Desktop/Sleep Data 128ch/s240_EMs_dir_-21_80ms_40mV_psd_10s-EM_density.csv'
# data_save = np.asarray(em_density_gruped)
# np.savetxt(fname_np_array,data_save,delimiter=',')
#
# ##### load test ###############
# test_np = np.loadtxt(fname_np_array,delimiter=',')