import mne
import os

from mne.datasets import fetch_fsaverage

# Download fsaverage files
fs_dir = fetch_fsaverage(verbose=True)
subjects_dir = os.path.dirname(fs_dir)

# The files live in:
subject = 'fsaverage'
trans = 'fsaverage'  # MNE has a built-in fsaverage transformation
src = os.path.join(fs_dir, 'bem', 'fsaverage-ico-5-src.fif')
bem = os.path.join(fs_dir, 'bem', 'fsaverage-5120-5120-5120-bem-sol.fif')


name_subject = 's248'
file_name = 's0248sleep.bdf'


data_path = os.path.expanduser("C:/Users/oledr/Desktop/Sleep Data 128ch/")
raw_fname = os.path.join(data_path,file_name)
raw = mne.io.read_raw_bdf(raw_fname, preload=False) # False
sfreq = raw.info['sfreq']

################### SETUP CHANNEL NAME AND MONTAGE ###################
channel_name = ['A1', 'A2', 'A3', 'A4', 'A5', 'A6', 'A7', 'A8', 'A9', 'A10',  'A11','A12', 'A13', 'A14', 'A15', 'A16', 'A17', 'A18', 'A19', 'A20', 'A21','A22', 'A23', 'A24', 'A25', 'A26', 'A27', 'A28', 'A29', 'A30', 'A31','A32', 'B1', 'B2', 'B3', 'B4', 'B5', 'B6','B7', 'B8', 'B9', 'B10','B11', 'B12', 'B13', 'B14', 'B15', 'B16', 'B17', 'B18', 'B19', 'B20','B21', 'B22', 'B23', 'B24', 'B25', 'B26', 'B27', 'B28', 'B29', 'B30','B31', 'B32', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9','C10', 'C11', 'C12', 'C13', 'C14', 'C15', 'C16', 'C17', 'C18', 'C19','C20', 'C21', 'C22', 'C23', 'C24', 'C25', 'C26', 'C27', 'C28', 'C29','C30', 'C31', 'C32', 'D1', 'D2', 'D3', 'D4', 'D5', 'D6', 'D7', 'D8','D9', 'D10', 'D11', 'D12', 'D13', 'D14', 'D15', 'D16', 'D17', 'D18','D19', 'D20', 'D21', 'D22', 'D23', 'D24', 'D25', 'D26', 'D27', 'D28','D29', 'D30', 'D31', 'D32']

mapping = dict()
for i, name in enumerate(channel_name):
    mapping[raw.ch_names[i]] = name
mne.rename_channels(raw.info, mapping)                                                                                          #}#
channel_dict = {'EMG-c':'emg','EMG-l':'emg','EMG-r':'emg','EOG-u': 'eog','EOG-l':'eog','EOG-r': 'eog','M1': 'misc', 'M2': 'misc'}#, 'GSR1': 'misc', 'GSR2': 'misc', 'Erg1': 'misc', 'Erg2': 'misc', 'Resp': 'misc', 'Plet': 'misc', 'Temp': 'misc'}
raw.set_channel_types(channel_dict)
# s129 no bad channels
#raw.drop_channels(['A31']) # s240
#raw.drop_channels(['B25','B26','B28','C30']) # s241
#raw.drop_channels(['A17','B3','B4','B13','C1','C31','D1','D18','D20','D21','D24']) #s242
#raw.drop_channels(['B19']) # s243
#raw.drop_channels(['A1','A2','B1','B3']) # s247
raw.drop_channels(['A1']) # s248
#raw.drop_channels(['A31','A32','B3','B19']) # s300

montage = mne.channels.make_standard_montage('biosemi128', head_size=0.095)
raw.set_montage(montage)
mne.set_eeg_reference(raw,ref_channels='average',projection=True)#['M1'])
################### SETUP CHANNEL NAME AND MONTAGE END ###################


fwd = mne.make_forward_solution(raw.info, trans=trans, src=src,
                                bem=bem, eeg=True, mindist=5.0, n_jobs=1)


noise_cov = mne.compute_raw_covariance(raw,tmin=2000, tmax=6000)


inverse_operator = mne.minimum_norm.make_inverse_operator(raw.info, fwd, noise_cov,
                                         loose=0.2, depth=0.8)

name_inverse = '{}-inv.fif'.format(name_subject)
fname_inverse = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/'.format(name_subject) + name_inverse
mne.minimum_norm.write_inverse_operator(fname_inverse,inverse_operator)
# inverse_operator = mne.minimum_norm.read_inverse_operator(fname_inverse)


############# LOAD EPOCHS ######################
name = '{}-epochs.fif'.format(name_subject)
fname_epochs = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/'.format(name_subject) + name
epochs = mne.read_epochs(fname_epochs)
evoked = epochs.average()
#evoked.drop_channels(['A31']) # s240
#evoked.drop_channels(['C1','C31','D21','D1']) # s242
evoked.plot()

############# LOAD EPOCHS END ######################


method = 'sLORETA' # 'sloreta' "dSPM" “MNE” | “dSPM” | “sLORETA” | “eLORETA”
snr = 3.
lambda2 = 1. / snr ** 2
stc = mne.minimum_norm.apply_inverse(evoked, inverse_operator, lambda2,
                    method=method, pick_ori=None)

stcs = mne.minimum_norm.apply_inverse_epochs(epochs, inverse_operator, lambda2, method,
                            pick_ori=None, return_generator=False)




########################## SAVE AND LOAD SOURCE ESTIMATES (STC) #######################
name_stc = '{}-stc-sLORETA.fif'.format(name_subject)
fname_stc = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/'.format(name_subject) + name_stc
stc.save(fname_stc)
stc_test = mne.read_source_estimate(fname_stc)
stc_test.plot('fsaverage')


for i, source in enumerate(stcs):
    name_stcs = '{}-{}-stcs.fif'.format(name_subject,i)
    fname_stcs = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/stcs from epochs/'.format(name_subject) + name_stcs
    source.save(fname_stcs)
    print('{} / {}'.format(i,len(stcs)))


stcs_test = []
for i in range(len(stcs)):
    name_stcs = '{}-{}-stcs.fif'.format(name_subject, i)
    fname_stcs = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/stcs from epochs/'.format(name_subject) + name_stcs
    stcs_test.append(mne.read_source_estimate(fname_stcs))
    print('{} / {}'.format(i, len(stcs)))

########################## SAVE AND LOAD SOURCE ESTIMATES (STC)  END #######################

