import mne
import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import time as timer
import datetime
start_time = timer.time()

sample_rate = 300  # sfreq
sfreq = 1024
################  Weighted Movment average Filter #################
def weighted_moving_average(data,n=int((7/80)*sample_rate)):
    weights = np.arange(1, n + 1)

    filter_data = pd.DataFrame(data)
    wmaN = filter_data.rolling(n).apply(lambda value: np.dot(value, weights) / weights.sum(), raw=True)

    return wmaN.values[:, 0]

def exp_weighted_moving_average(data,n=int((7/80)*sample_rate/2)):
    filter_data = pd.DataFrame(data)
    ema10 = filter_data.ewm(span=n).mean()
    #wmaN = filter_data.rolling(n).apply(lambda value: np.dot(value, weights) / weights.sum(), raw=True)

    return np.array(ema10[0])


#################### LOW PASS FILTER ############################
from scipy.signal import butter,filtfilt
# Filter requirements.
fs = sample_rate       # sample rate, Hz
cutoff = 200      # desired cutoff frequency of the filter, Hz ,      slightly higher than actual 1.2 Hz
nyq = 0.5 * fs  # Nyquist Frequency
order = 8       # sin wave can be approx represented as quadratic


import samplerate
sample_ratio = sample_rate / sfreq
converter = 'sinc_best'  # or 'sinc_fastest', ...


def butter_lowpass_filter(data, cutoff, fs, order):
    normal_cutoff = cutoff / nyq
    # Get the filter coefficients
    b, a = butter(order, normal_cutoff, btype='low', analog=False)
    y = filtfilt(b, a, data)
    return y




def D(xlist,ylist):
    yprime = np.diff(ylist)/np.diff(xlist)
    xprime = []
    for i in range(len(yprime)):
        xstep = (xlist[i+1]+xlist[i])/2
        xprime = np.append(xprime,xstep)
    return xprime, yprime


def Amplitude_finder(data,n=int((30/80)*sample_rate)):
    # Looking at 40 (0.5s) to see if value change is greater then 2*30 micro volt as
    # Should change to looking at smaller time then 0.5 sec also !!!!!!
    data_frame = pd.DataFrame(data)
    value_change = data_frame.rolling(n).apply(lambda value: abs(value[0:4].mean()-value[-5:-1].mean()), raw=True)
    index = np.where(abs(value_change) > 2*30/10**6)
    index_fixed = np.array([x - int(n) for x in index[0]])
    return index_fixed



def remove_lonly_elements(arr,n=int((5/80)*sample_rate)):
    # Removes elements with less then n elements close to it
    # Not in use anymore
    index_set = set()
    for i in range(len(arr) - n):
        if arr[i + n] - arr[i] <= n + 1:
            for value in arr[i:i + n + 1]:
                # print(arr[i:i+n])
                index_set.add(value)
    index_list = list(index_set)
    index_list.sort()
    return index_list

def ok_slope_average(arr,n=int((10/80)*sample_rate)):
    # Check for where mean of next n values are over criteria and where the fist value is over the criteria
    criteria_value = 2 * 250 / 10 ** 6
    data_frame = pd.DataFrame(arr)
    # find the mean of the slope over n consecutive values
    value_change = data_frame.rolling(n).apply(lambda value: np.mean(value), raw=True)
    # move the array n spots so pos 0 = mean for 0 - n
    arr_mean_value = np.array(value_change[n-1:-1][0])
    arr_out = []
    # find indexes for where the value are greater then criteria
    index_value_over = np.where(abs(arr) > criteria_value)
    # check for mean over next n points is over criteria and if start point is over criteria
    for i in range(len(arr_mean_value)):
        if abs(arr_mean_value[i]) > criteria_value :
            if i in index_value_over[0]:
                arr_out = np.append(arr_out,i)
    # return indexes of points that meet both criterias
    return np.array(arr_out,int)

def Slope_finder(data_prime,n=int((40/80)*sample_rate)):
    # Finding indexes of points where data dirivative is greater then 2*248.3 micro volts
    #points_value = data_prime[np.where(data_prime > 2*248.3/10**6)]
    #point_index = np.where(abs(data_prime) > 2*248.3/10**6 )

    # Only accept where its n point in a row
    #indexes = remove_lonly_elements(point_index[0])
    # Only accept when average slope over 10 points is over 2*248.3
    indexes = ok_slope_average(data_prime)
    return indexes



def find_index_2prime(EMs_points,y2prime,data,n=int((5/80)*sample_rate)):
    # Finding maximum of double derivative function close to the points
    # This should give the start of the peak
    # Make sure the maximum is in the right direction (if value goes down it needs to be the minimum of the y2prime)
    indexes_EMs = []
    for value in EMs_points:
        value = int(value)
        if value <= n:
            indexes_EMs.append(value)
            continue
        if data[value] < data[value+2*n]:
            index = np.argmax(y2prime[value - n:value+n]) # value
        else:
            index = np.argmin(y2prime[value - n:value+n]) # value
        indexes_EMs.append(value-n+index)

    indexes_EMs = np.array(indexes_EMs)
    return indexes_EMs

def remove_close_elements(arr):
    # Remove elements to close to approved element. to close is definded by closness
    # there needs to be a gap bigger then closeness between point a and b for the point b to be accepted
    remove_after = set()
    closeness = 0.2 * sample_rate # 0.2 sec
    for i in range(len(arr)-1):
        if abs(arr[i+1] - arr[i]) < closeness:
            remove_after.add(i+1)
    if len(remove_after) < 1:
        return arr
    new_arr = np.delete(arr, np.array(list(remove_after)))
    return new_arr.astype(int)

def find_EMs(data_double_prime,slopes,amplitudes,data):
    # Finding indexes where both the slope and amplitude is criteria is meet
    shared_values = []
    for value in slopes:
        if value in amplitudes:
            shared_values = np.append(shared_values, value)
    if len(shared_values) < 1:
         return []

    return_array = remove_close_elements(shared_values)
    # if len(return_array) < 1:
    #     return []
    # move point to where the double derivative is larges, should be at the start of a slope
    hold = find_index_2prime(return_array,data_double_prime,data)
    hold = set(hold)
    hold = np.array(list(hold))
    return hold


def check_opposite_spike(EM_points,EOGl,EOGr):
    n = int(0.2*sample_rate)
    criteria = 30/10**6 # 30 til 45 til 55
    point_ok = []
    for value in EM_points:
        left_spike = np.mean(EOGl[value+n-int(n/20):value+n+int(n/20)]) - np.mean(EOGl[value-int(n/20):value+int(n/20)])
        right_spike = np.mean(EOGr[value+n-int(n/20):value+n+int(n/20)]) - np.mean(EOGr[value-int(n/20):value+int(n/20)])
        if (left_spike > criteria and right_spike < criteria) or (left_spike <criteria and right_spike > criteria):
            point_ok = np.append(point_ok,value)
    return np.array(point_ok,int)


def check_induvidual_channels(EM_points,EOG_left,EOG_right):
    step = int(0.08*sample_rate) # = 50 milliseconds
    n = 4 # for å ungå spike points
    first_step = int(0.021*sample_rate)
    criteria = 14/10**6 # 14 as default
    point_ok = []
    for em in EM_points:
        right_em_spike = 0
        left_em_spike = 0
        for i in range(1,4):
            if i == 1:
                left_spike = np.mean(EOG_left[em - n + step:em + n + step]) - np.mean(EOG_left[em - n - first_step:em + n + -first_step])
                right_spike = np.mean(EOG_right[em - n + step:em + n + step]) - np.mean(EOG_right[em - n - first_step:em + n + -first_step])
            # else:
            #     left_spike = np.mean(EOG_left[em-n+step*(1+i):em+n+step*(1+i)]) -np.mean(EOG_left[em-n+step*i:em+n+step*i])
            #     right_spike = np.mean(EOG_right[em-n+step*(1+i):em+n+step*(1+i)]) - np.mean(EOG_right[em-n+step*i:em+n+step*i])
            if (left_spike > criteria and right_spike < -criteria):
                right_em_spike += 1
            if (left_spike < -criteria and right_spike > criteria):
                left_em_spike += 1
        if left_em_spike >= 1 or right_em_spike >= 1:
            point_ok = np.append(point_ok,em)
    return np.array(point_ok,int)
# raw = raw_full_size.copy()
# raw.crop(tmin=actual_min,tmax=actual_max)
# raw.plot(events=events_found)


EOG_l_channel_name = 'EOG-l'   # 'EOG-l'  'EXG7'
EOG_r_channel_name = 'EOG-r'  # 'EOG-r'  'EXG8'
EMG_c_channel_name = 'EMG-c' #,'EMG-l','EMG-r'])
# raw_full_size.pick_channels([EOG_l_channel_name, EOG_r_channel_name,'M2'])
# raw_full_size.load_data()
# mne.set_eeg_reference(raw_full_size,['M2'])
# raw_full_size.notch_filter(50)
# fmin, fmax = 0.5, 150
# raw_full_size.filter(fmin, fmax)

subject_name = 's300' # this and rem_namelist needs to be change, nothing else
#rem_namelist = ['s0129-REM1-1.bdf','s0129-REM2-1.bdf','s0129-REM3-1.bdf','s0129-REM4-1.bdf','s0129-REM4-2.bdf','s0129-REM4-3.bdf','s0129-REM4-4.bdf']
#rem_namelist = ['s0240adaptation-REM1-1.bdf','s0240adaptation-REM1-2.bdf','s0240adaptation-REM2-1.bdf','s0240adaptation-REM2-2.bdf','s0240adaptation-REM3-1.bdf','s0240adaptation-REM4-1.bdf','s0240adaptation-REM4-2.bdf','s0240adaptation-REM5-1.bdf']
#rem_namelist = ['s0241REM1-1.bdf','s0241REM2-1.bdf','s0241REM3-1.bdf','s0241REM3-2.bdf','s0241REM3-3.bdf','s0241REM4-1.bdf']
#rem_namelist = ['s0242adaptation-REM1-1.bdf','s0242adaptation-REM2-1.bdf','s0242adaptation-REM3-2.bdf','s0242adaptation-REM4-1.bdf','s0242adaptation-REM4-2.bdf','s0242adaptation-REM5-1.bdf','s0242adaptation-REM5-2.bdf','s0242adaptation-REM5-3.bdf','s0242adaptation-REM5-4.bdf' ]
#rem_namelist = ['s0243-REM1-1.bdf','s0243-REM1-2.bdf','s0243-REM2-1.bdf','s0243-REM2-2.bdf','s0243-REM3-1.bdf','s0243-REM3-2.bdf','s0243-REM3-3.bdf','s0243-REM3-4.bdf','s0243-REM3-5.bdf','s0243-REM3-6.bdf']
#rem_namelist = ['s0247adaptation-REM1-1.bdf','s0247adaptation-REM1-2.bdf','s0247adaptation-REM1-3.bdf','s0247adaptation-REM1-4.bdf','s0247adaptation-REM1-5.bdf','s0247adaptation-REM1-6.bdf','s0247adaptation-REM2-1.bdf','s0247adaptation-REM2-2.bdf','s0247adaptation-REM2-3.bdf','s0247adaptation-REM2-4.bdf','s0247adaptation-REM2-5.bdf','s0247adaptation-REM3-1.bdf','s0247adaptation-REM3-2.bdf','s0247adaptation-REM3-3.bdf','s0247adaptation-REM3-4.bdf','s0247adaptation-REM3-5.bdf','s0247adaptation-REM3-6.bdf','s0247adaptation-REM3-7.bdf','s0247adaptation-REM3-8.bdf']
#rem_namelist = ['s0248-REM1-1.bdf','s0248-REM1-2.bdf','s0248-REM2-1.bdf','s0248-REM2-2.bdf','s0248-REM2-3.bdf','s0248-REM3-1.bdf','s0248-REM3-2.bdf','s0248-REM3-3.bdf','s0248-REM4-1.bdf','s0248-REM4-2.bdf','s0248-REM4-3.bdf','s0248-REM4-4.bdf','s0248-REM5-1.bdf','s0248-REM5-2.bdf']
rem_namelist = ['s0300-REM1-1.bdf','s0300-REM1-2.bdf','s0300-REM1-3.bdf','s0300-REM1-4.bdf','s0300-REM2-1.bdf','s0300-REM2-2.bdf','s0300-REM2-3.bdf','s0300-REM3-1.bdf','s0300-REM3-2.bdf']
for name in rem_namelist:

    data_path = os.path.expanduser("C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/".format(subject_name))

    raw_fname = os.path.join(data_path,name)
    raw_full_size = mne.io.read_raw_bdf(raw_fname, preload=False) # False
    sfreq = raw_full_size.info['sfreq']
    tmin, tmax = 0, len(raw_full_size)/sfreq-0.1



    ########## For å kjøre lengre deler  ##################
    time_per_loop = 100
    nr_of_runs_needed = (tmax-tmin)/time_per_loop
    actual_max = tmax
    actual_min = tmin
    EM_points_saved = []
    EMG_c_value = []
    print('Running algo on tmin = {}, tmax = {} , number of runs = {}'.format(actual_min,actual_max,nr_of_runs_needed))
    for long_run in range(int(nr_of_runs_needed)+1):
        loop_start = timer.time()
        if long_run > int(nr_of_runs_needed)-1:
            tmin = actual_min+long_run * time_per_loop
            tmax = actual_max
        else:
            tmin = actual_min+long_run*time_per_loop
            tmax = tmin+time_per_loop

        raw = raw_full_size.copy()
        raw = raw.crop(tmin=tmin, tmax=tmax)
        raw.load_data()
        # alt må inn her!

        #raw.notch_filter(50)


        sfreq = raw_full_size.info['sfreq']
        time = raw.pick_channels([EOG_l_channel_name, EOG_r_channel_name]).times
        data = raw.pick_channels([EOG_l_channel_name, EOG_r_channel_name])._data
        scalar = 10 ** 6


        EOG_left = data[0]
        EOG_right = data[1]
        # data = np.subtract(weighted_moving_average(data[1]),weighted_moving_average(data[0]))
        data_samlet = np.subtract(EOG_right, EOG_left)

        data_sampled = samplerate.resample(data_samlet, sample_ratio, converter)
        time_sampled = samplerate.resample(time, sample_ratio, converter)


        time = time_sampled

        data = weighted_moving_average(data_sampled)


        EOG_left = samplerate.resample(EOG_left, sample_ratio, converter)
        EOG_right = samplerate.resample(EOG_right, sample_ratio, converter)

        ################### Derivasjon ##########################

        xprime, yprime = D(time, data)

        x2prime, y2prime = D(xprime, yprime)


        ################ KRITERIA ##################################

        # Amplitude greater then 30 microvolts = betyr  10**6*volt endring = 30

        # Duration less then 0.5 secounds = betyr mindre en 40 steps i arrayet

        # Slope greater then 248.3 microvolts / secound = derivering større en 248.3/10**6 = 0.0002483

        amplitude_marker = Amplitude_finder(data)  ## legg in litt og litt loop

        derivert_marker = Slope_finder(yprime)  ## legg in litt og litt loop

        EM_points = find_EMs(y2prime, derivert_marker, amplitude_marker, data)  ## legg in litt og litt loop

        EM_points_checked = check_opposite_spike(EM_points, EOG_left, EOG_right)  ## legg in litt og litt loop
        EM_points_checked2 = check_induvidual_channels(EM_points, EOG_left, EOG_right)

        print('EM gammel spike check = {}, EM new spike check {}'.format(len(EM_points_checked),len(EM_points_checked2)))
        ################## Make EM points events for data ##################################
        new_events = np.zeros(shape=(len(EM_points_checked2), 3), dtype=int)

        event_time = [time[x] for x in EM_points_checked2]
        event_time.sort()
        event_time = [x + tmin for x in event_time]
        event_time_new = np.array([int(x * sample_rate / sample_ratio) for x in event_time])

        new_events[:, 0] = event_time_new
        new_events[:, 2] = np.ones(len(event_time_new))

        EM_points_saved.append(new_events)



        loop_end = timer.time()
        total_secound_in_loop = loop_end - start_time
        loop_seounds = loop_end - loop_start
        loop_time = str(datetime.timedelta(seconds=loop_seounds))
        total_time_in_loop = str(datetime.timedelta(seconds=total_secound_in_loop))
        print('Total time of loop {}/{} = {}, total = {}'.format(long_run,int(nr_of_runs_needed),loop_time,total_time_in_loop))
        print('Loop nr {}/{}, tmin = {}, tmax = {}'.format(long_run,int(nr_of_runs_needed),tmin,tmax))
        print('Nr of events in this loop = {}'.format(len(EM_points_checked2)))



    end_time = timer.time()
    total_secound = end_time - start_time

    total_time = str(datetime.timedelta(seconds=total_secound))

    print('Total time of the script = {}'.format(total_time))

    ########################## Reformat events to standard event format ######################################
    total_event_lenght = 0
    for arr in EM_points_saved:
        total_event_lenght += len(arr)
    events_found = np.zeros(shape=(total_event_lenght, 3), dtype=int)

    length_to_now = 0
    for i, arr in enumerate(EM_points_saved):
        events_found[length_to_now:length_to_now+len(EM_points_saved[i])] = arr
        length_to_now += len(EM_points_saved[i])
    print('nr of events found = {}'.format(len(events_found[:,0])))
    events_per_min = len(events_found[:,0])*60/actual_max
    print('Events per minute = {}'.format(events_per_min))

    ########################## SAVE EVENTS TO FIF FILE ##################################

    event_new_fname = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/events/{}-no-psd-eve.fif'.format(subject_name,name[0:-4])
    mne.write_events(event_new_fname,events_found)

    ########################## SAVE EVENTS TO FIF FILE ##################################

    event = events_found
    start_stop = []
    time_span = 5
    for i, time in enumerate(event[:, 0]):
        # find tmin and tmax
        if time < time_span * 1024 or int(time / 1024) > tmax - time_span:
            start_stop.append([0, 0])
            continue
        start_stop.append([int(time / 1024 - time_span), int(time / 1024 + time_span)])

    meanvalue = []
    for i, time in enumerate(start_stop):
        if time[0] < 10:
            meanvalue.append(0)
            continue
        # if i < 1000 or i > 1500:
        #     continue
        raw = raw_full_size.copy()
        raw.pick(['EMG-c'])  # ,'EMG-r','EMG-l'])
        raw.crop(tmin=time[0], tmax=time[1])
        raw.load_data()
        fmin, fmax = 55, 95
        # raw.plot_psd(fmin=fmin,fmax=fmax)
        pas, freqs = mne.time_frequency.psd_array_welch(raw._data, sfreq, fmin=fmin, fmax=fmax, n_fft=1024 * 2,
                                                        average='mean')

        pas[0] = pas[0] * 10 ** 12
        pas[0] = 10 * np.log10(pas[0])  # This to convert to dB
        # plt.plot(freqs,pas[0],'--')
        meanvalue.append(sum(pas[0]) / len(pas[0]))
        print('RUN NR {} / {}'.format(i, len(start_stop)))

    # plt.plot(meanvalue)
    # meanvaluehold
    # meanvalue.extend(meanvaluehold)

    new_EMs = []
    for i, value in enumerate(meanvalue):
        if value < -11:
            new_EMs.append(event[i, 0])

    events_found = np.zeros(shape=(len(new_EMs), 3), dtype=int)

    events_found[:, 0] = new_EMs
    events_found[:, 2] = np.ones(len(new_EMs), dtype=int)
    print('EMs foud = ',len(new_EMs))
    ########################## SAVE EVENTS TO FIF FILE ##################################

    event_new_fname = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/events/{}-eve.fif'.format(subject_name,name[0:-4])
    mne.write_events(event_new_fname,events_found)

    raw = raw_full_size.copy()
    EOG_l_channel_name = 'EOG-l'  # 'EOG-l'  'EXG7'
    EOG_r_channel_name = 'EOG-r'  # 'EOG-r'  'EXG8'
    raw.pick_channels([EOG_l_channel_name])
    raw.load_data()

    ##### Add EM channel
    stim_data = np.zeros((1, len(raw.times)))
    info = mne.create_info(['EM'], sfreq, ['stim'])
    stim_raw = mne.io.RawArray(stim_data, info)
    raw.add_channels([stim_raw], force_update_info=True)
    raw.add_events(events_found, 'EM')
    raw.pick_channels(['EM'])

    # raw.plot(events=event)

    ##### Save Raw
    fname_rem = 'C:/Users/oledr/Desktop/Sleep Data 128ch/REM only/{}/EM channel raw/{}-EM-channel-only-raw.fif'.format(subject_name,name[0:-4])
    raw.save(fname_rem)